#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Following code fixes 2 issues with kafka-python and
# 1. >> The current release of eventlet (0.19.0) does not actually remove
# select.poll [1]. Because of kafka-python.selectors34 selects
# PollSelector instead of SelectSelector [2]. PollSelector relies on
# select.poll, which does not work when eventlet/greenlet is used. This
# bug in evenlet is fixed in the master branch [3], but there's no
# release of eventlet that includes this fix at this point.
# 2. >>kafka.selectors34.SelectSelector does not actually work when its
# _select function is actually the green version provided by eventlet.
# It does not work because _select is called using self._select(...),
# implying that the instance (self) is passed to "_select" as the first
# parameter. The green version of select expects 4 arguments, and it is
# called with 5.

try:
    from eventlet import patcher
    if patcher.is_monkey_patched('select'):
        import eventlet.green.select
        import kafka.selectors34

        # make SelectSelector._select a staticmethod as done in
        # itself
        # https://github.com/eventlet/eventlet/blob/master/eventlet/green/
        # selectors.py#L32
        kafka.selectors34.SelectSelector._select = staticmethod(
            eventlet.green.select.select)
        # eventlet <= 1.9.0 does not remove select.poll, so we manually set
        # DefaultSelector to SelectSelector
        # see https://github.com/eventlet/eventlet/pull/317
        kafka.selectors34.DefaultSelector = kafka.selectors34.SelectSelector
except ImportError:
    pass

import json
import threading

from oslo_messaging._drivers import base
from oslo_messaging._drivers import common as driver_common
from oslo_messaging._drivers import pool as driver_pool
from oslo_messaging._i18n import _LE
from oslo_messaging._i18n import _LI
from oslo_messaging._i18n import _LW
from oslo_serialization import jsonutils

import kafka
import kafka.common
from oslo_config import cfg
from oslo_log import log as logging
import retrying

LOG = logging.getLogger(__name__)

PURPOSE_SEND = 'send'
PURPOSE_LISTEN = 'listen'

DEFAULT_RECONNECT_DELAY = 1000


kafka_opts = [
    cfg.StrOpt('kafka_default_host', default='localhost',
               deprecated_for_removal=True,
               deprecated_reason="Replaced by [DEFAULT]/transport_url",
               help='Default Kafka broker Host'),

    cfg.PortOpt('kafka_default_port', default=9092,
                deprecated_for_removal=True,
                deprecated_reason="Replaced by [DEFAULT]/transport_url",
                help='Default Kafka broker Port'),

    cfg.IntOpt('kafka_max_fetch_bytes', default=1024 * 1024,
               help='Max fetch bytes of Kafka consumer'),

    cfg.IntOpt('kafka_consumer_timeout', default=1.0,
               help='Default timeout(s) for Kafka consumers'),

    cfg.IntOpt('pool_size', default=10,
               help='Pool Size for Kafka driver'),

    cfg.StrOpt('consumer_group', default="oslo_messaging_consumer",
               help='Group id for Kafka consumer. Consumers in one group '
                    'will coordinate message consumption'),

    cfg.FloatOpt('producer_batch_timeout', default=0.,
                 help="Upper bound on the delay for KafkaProducer batching "
                      "in seconds"),

    cfg.IntOpt('producer_batch_size', default=16384,
               help='Size of batch for the producer async send')
]

CONF = cfg.CONF


def unpack_message(msg):
    context = {}
    message = None
    try:
        if msg:
            msg = json.loads(msg)
            message = driver_common.deserialize_msg(msg)
            if 'context' in message:
                context = message['context']
                del message['context']
    except ValueError as e:
        LOG.info(_LI("Invalid format of consumed message: %s"), e)
    except Exception as e:
        LOG.warning(_LW("Exception during message unpacking: %s"), e)
    return message, context


def pack_message(ctxt, msg):
    """Pack context into msg."""

    if isinstance(ctxt, dict):
        context_d = ctxt
    else:
        context_d = ctxt.to_dict()
    msg['context'] = context_d

    msg = driver_common.serialize_msg(msg)

    return msg


def target_to_topic(target, priority=None):
    """Convert target into topic string

    :param target: Message destination target
    :type target: oslo_messaging.Target
    :param priority: Notification priority
    :type priority: string
    """
    if not priority:
        return target.topic
    return target.topic + '.' + priority


def retry_on_retriable_kafka_error(err):
    return (isinstance(err, kafka.common.KafkaError) and
            getattr(err, "retriable", False))


def with_reconnect(retries=None, delay=None):
    def decorator(func):
        @retrying.retry(
            retry_on_exception=retry_on_retriable_kafka_error,
            wait_fixed=delay or DEFAULT_RECONNECT_DELAY,
            stop_max_attempt_number=retries
        )
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return decorator


class Producer(object):
    _producer = None
    _servers = None
    _lock = threading.Lock()

    @staticmethod
    @with_reconnect()
    def connect(servers, **kwargs):
        return kafka.KafkaProducer(
            bootstrap_servers=servers, **kwargs)

    @classmethod
    def producer(cls, servers, **kwargs):
        with cls._lock:
            if not cls._producer or cls._servers != servers:
                cls._servers = servers
                cls._producer = cls.connect(servers, **kwargs)
        return cls._producer

    @classmethod
    def cleanup(cls):
        with cls._lock:
            if cls._producer:
                cls._producer.close()
            cls._producer = None


class Connection(object):

    def __init__(self, conf, url, purpose):

        self.client = None
        driver_conf = conf.oslo_messaging_kafka
        self.batch_size = driver_conf.producer_batch_size
        self.linger_ms = driver_conf.producer_batch_timeout * 1000
        self.conf = conf
        self.producer = None
        self.consumer = None
        self.consumer_timeout = float(driver_conf.kafka_consumer_timeout)
        self.max_fetch_bytes = driver_conf.kafka_max_fetch_bytes
        self.group_id = driver_conf.consumer_group
        self.url = url
        self._parse_url()
        # TODO(Support for manual/auto_commit functionality)
        # When auto_commit is False, consumer can manually notify
        # the completion of the subscription.
        # Currently we don't support for non auto commit option
        self.auto_commit = True
        self._consume_loop_stopped = False

    def _parse_url(self):
        driver_conf = self.conf.oslo_messaging_kafka
        self.hostaddrs = []

        for host in self.url.hosts:
            if host.hostname:
                self.hostaddrs.append("%s:%s" % (
                    host.hostname,
                    host.port or driver_conf.kafka_default_port))

        if not self.hostaddrs:
            self.hostaddrs.append("%s:%s" % (driver_conf.kafka_default_host,
                                             driver_conf.kafka_default_port))

    def notify_send(self, topic, ctxt, msg, retry):
        """Send messages to Kafka broker.

        :param topic: String of the topic
        :param ctxt: context for the messages
        :param msg: messages for publishing
        :param retry: the number of retry
        """

        message = pack_message(ctxt, msg)
        self._ensure_connection()
        self._send_and_retry(message, topic, retry)

    def _send_and_retry(self, message, topic, retry):
        if not isinstance(message, str):
            message = jsonutils.dumps(message)
        retry = retry if retry >= 0 else None

        @with_reconnect(retries=retry)
        def _send(topic, message):
            self.producer.send(topic, message)

        try:
            _send(topic, message)
        except Exception:
            Producer.cleanup()
            LOG.exception(_LE("Failed to send message"))

    @with_reconnect()
    def _poll_messages(self, timeout):
        return self.consumer.poll(timeout)

    def consume(self, timeout=None):
        """Receive up to 'max_fetch_messages' messages.

        :param timeout: poll timeout in seconds
        """
        if self._consume_loop_stopped:
            return None

        timeout = timeout if timeout >= 0 else self.consumer_timeout
        try:
            messages = self._poll_messages(timeout)
        except kafka.common.ConsumerTimeout as e:
            raise driver_common.Timeout(e.message)
        except Exception:
            LOG.exception(_LE("Failed to consume messages"))
            messages = None
        return messages

    def stop_consuming(self):
        self._consume_loop_stopped = True

    def reset(self):
        """Reset a connection so it can be used again."""
        pass

    def close(self):
        if self.producer:
            self.producer.close()
        self.producer = None
        if self.consumer:
            self.consumer.close()
        self.consumer = None

    def commit(self):
        """Commit is used by subscribers belonging to the same group.

        After subscribing messages, commit is called to prevent
        the other subscribers which belong to the same group
        from re-subscribing the same messages.

        Currently self.auto_commit option is always True,
        so we don't need to call this function.
        """
        self.consumer.commit()

    def _ensure_connection(self):
        try:
            self.producer = Producer.producer(self.hostaddrs,
                                              linger_ms=self.linger_ms,
                                              batch_size=self.batch_size)
        except kafka.common.KafkaError as e:
            LOG.exception(_LE("KafkaProducer could not be initialized: %s"), e)
            raise

    @with_reconnect()
    def declare_topic_consumer(self, topics):
        self.consumer = kafka.KafkaConsumer(
            *topics, group_id=self.group_id,
            bootstrap_servers=self.hostaddrs,
            max_partition_fetch_bytes=self.max_fetch_bytes
        )


class OsloKafkaMessage(base.RpcIncomingMessage):

    def __init__(self, ctxt, message):
        super(OsloKafkaMessage, self).__init__(ctxt, message)

    def requeue(self):
        LOG.warning(_LW("requeue is not supported"))

    def reply(self, reply=None, failure=None):
        LOG.warning(_LW("reply is not supported"))


class KafkaListener(base.Listener):

    def __init__(self, conn):
        super(KafkaListener, self).__init__()
        self._stopped = threading.Event()
        self.conn = conn
        self.incoming_queue = []

    @base.batch_poll_helper
    def poll(self, timeout=None):
        while not self._stopped.is_set():
            if self.incoming_queue:
                return self.incoming_queue.pop(0)
            try:
                messages = self.conn.consume(timeout=timeout)
                if messages:
                    self._put_messages_to_queue(messages)
            except driver_common.Timeout:
                return None

    def _put_messages_to_queue(self, messages):
        for topic, records in messages.items():
            if records:
                for record in records:
                    message, context = unpack_message(record.value)
                    if message:
                        self.incoming_queue.append(
                            OsloKafkaMessage(ctxt=context, message=message))

    def stop(self):
        self._stopped.set()
        self.conn.stop_consuming()

    def cleanup(self):
        self.conn.close()

    def commit(self):
        # TODO(Support for manually/auto commit functionality)
        # It's better to allow users to commit manually and support for
        # self.auto_commit = False option. For now, this commit function
        # is meaningless since user couldn't call this function and
        # auto_commit option is always True.
        self.conn.commit()


class KafkaDriver(base.BaseDriver):
    """Note: Current implementation of this driver is experimental.

    We will have functional and/or integrated testing enabled for this driver.
    """

    def __init__(self, conf, url, default_exchange=None,
                 allowed_remote_exmods=None):

        opt_group = cfg.OptGroup(name='oslo_messaging_kafka',
                                 title='Kafka driver options')
        conf.register_group(opt_group)
        conf.register_opts(kafka_opts, group=opt_group)

        super(KafkaDriver, self).__init__(
            conf, url, default_exchange, allowed_remote_exmods)

        self.connection_pool = driver_pool.ConnectionPool(
            self.conf, self.conf.oslo_messaging_kafka.pool_size,
            self._url, Connection)
        self.listeners = []

    def cleanup(self):
        for c in self.listeners:
            c.close()
        self.listeners = []

    def send(self, target, ctxt, message, wait_for_reply=None, timeout=None,
             retry=None):
        raise NotImplementedError(
            'The RPC implementation for Kafka is not implemented')

    def send_notification(self, target, ctxt, message, version, retry=None):
        """Send notification to Kafka brokers

        :param target: Message destination target
        :type target: oslo_messaging.Target
        :param ctxt: Message context
        :type ctxt: dict
        :param message: Message payload to pass
        :type message: dict
        :param version: Messaging API version (currently not used)
        :type version: str
        :param retry: an optional default kafka consumer retries configuration
                      None means to retry forever
                      0 means no retry
                      N means N retries
        :type retry: int
        """
        with self._get_connection(purpose=PURPOSE_SEND) as conn:
            conn.notify_send(target_to_topic(target), ctxt, message, retry)

    def listen(self, target):
        raise NotImplementedError(
            'The RPC implementation for Kafka is not implemented')

    def listen_for_notifications(self, targets_and_priorities, pool=None):
        """Listen to a specified list of targets on Kafka brokers

        :param targets_and_priorities: List of pairs (target, priority)
                                       priority is not used for kafka driver
                                       target.exchange_target.topic is used as
                                       a kafka topic
        :type targets_and_priorities: list
        :param pool: consumer group of Kafka consumers
        :type pool: string
        """
        conn = self._get_connection(purpose=PURPOSE_LISTEN)
        topics = set()
        for target, priority in targets_and_priorities:
            topics.add(target_to_topic(target, priority))

        conn.declare_topic_consumer(topics)

        listener = KafkaListener(conn)
        return listener

    def _get_connection(self, purpose):
        return driver_common.ConnectionContext(self.connection_pool, purpose)
